let serial
let data

let width = 800
let height = 800

///////////////////////// Create Scales /////////////////////////
const max_value = 1023
const colors = ['#423b8a', '#6a469b', '#9d4ba4', '#d744a6', '#ff587a', '#fc8c36', '#efb605'].reverse()
const scale_color = d3.scaleLinear()
    .domain(d3.range(colors.length).map(d => { return d/(colors.length-1) * max_value}))
    .range(colors)
    .clamp(true)

//Stroke width of the lines
const scale_width = d3.scaleLinear()
    .domain([0, max_value])
    .range([3, 0.3])

//Opacity of the lines
const scale_opacity = d3.scaleLinear()
    .domain([0, max_value])
    .range([255, 255])

///////////////////////// Create Agents /////////////////////////
let agents = []
const agent_num = 2000
const noise_scale = 100
const noise_strength = 10
const noise_z_range = 0.4
const noise_z_velocity = 0.01

let Agent = function(noise_z_range) {
    this.vector = createVector(random(width), random(height))
    this.vector_old = this.vector.copy()
    this.step_size = random(1, 5)
    this.angle
    this.noise_z = random(noise_z_range)
    this.color = "#423b8a"
}//Agent

//Based on: https://editor.p5js.org/generative-design/sketches/M_1_5_03
Agent.prototype.update = function(stroke_width, noise_scale, noise_strength, noise_z_velocity) {
    //Get the flow field angle from the Perlin noise field
    this.angle = noise(this.vector.x / noise_scale, this.vector.y / noise_scale, this.noise_z) * noise_strength
  
    //Add the new "step" to the current position, where step_size is basically the velocity
    this.vector.x += cos(this.angle) * this.step_size
    this.vector.y += sin(this.angle) * this.step_size
  
    //Check if the agent has run outside of the screen
    //If yes, move it to the other side
    if (this.vector.x < -10) this.vector.x = this.vector_old.x = width + 10
    if (this.vector.x > width + 10) this.vector.x = this.vector_old.x = -10
    if (this.vector.y < -10) this.vector.y = this.vector_old.y = height + 10
    if (this.vector.y > height + 10) this.vector.y = this.vector_old.y = -10
  
    //Set up the appearance and draw the tiny piece of path
    stroke(this.color)
    strokeWeight(stroke_width * this.step_size) //somewhat random stroke width
    line(this.vector_old.x, this.vector_old.y, this.vector.x, this.vector.y)
  
    //Put the current position into the "old" position for the next time update is called
    this.vector_old = this.vector.copy()
  
    //Move the z-axis along a little for each agent to get a new Perlin structure
    this.noise_z += noise_z_velocity
}//Agent.update

///////////////////////////// Setup /////////////////////////////
function setup() {
    // Instantiate our SerialPort object
    serial = new p5.SerialPort()
    // Assuming the board is connected, let's open the connection to it
    // Use the FTDI ID
    serial.open("/dev/tty.usbserial-D30A3T7C")
    // When we connect to the underlying server
    serial.on('connected', () => { console.log("We are connected!") })
    // When we some data from the serial port
    serial.on('data', gotData)

    /////////// Setup general p5.js stuff ///////////
    let canvas = createCanvas(width, height)
    canvas.parent("chart")

    //Create all agents
    for (let i = 0; i < agent_num; i++) {
        agents[i] = new Agent(noise_z_range)
    }//for i
}//function setup
  
////////////////////////////// Draw /////////////////////////////
function draw() {
    if (data) {
        //Adjust the text value on top
        document.getElementById("text-value").innerHTML = data

        // //Draw a very faint white background over the previous pass
        // background('rgba(255,255,255, 0.01)')
    
        //Update and draw each agent
        for (let i = 0; i < agent_num; i++) {
            //Adjust the color (sometimes)
            if(random() < 0.05) {
                let agent_color = colorOffset(scale_color(data))
                let rgb = d3.rgb(agent_color)
                let stroke_opacity = scale_opacity(data)
                agents[i].color = color(rgb.r, rgb.g, rgb.b, stroke_opacity)
            }//if

            //Get the stroke width
            let stroke_width = scale_width(data)

            //Draw the agent
            agents[i].update(stroke_width, noise_scale, noise_strength, noise_z_velocity)
        }//for i
    }//if
}//function draw

//////////////////// Serial Helper Functions ////////////////////
// There is data available to work with from the serial port
function gotData() {
    let inString = serial.readLine()

    //Check to see that there's actually a string there
    if (inString.length > 0) {
        //Convert it to a number
        data = Number(inString)
    }//if
}//function gotData

//////////////////////// Helper Functions ///////////////////////
//https://github.com/bgrins/TinyColor
//Get a slightly different color, based on the provided color
function colorOffset(color) {
	let colors = tinycolor(color).analogous( results = 3, slices = 15)
	colors = colors.map(t => t.toHexString())

	return colors[Math.floor(random() * colors.length)]
}//function colorOffset
